; 3 card solitaire
; a 1K ZX81 implementation of the famous cardgame
; WITH CHROMA COLOUR AND GRAPHICS SUPPORT!

; Suits
; S=Spades
; C=Clubs
; D=Diamonds
; H=Hearts

; Cards
; A=Ace
; 1 to 9
; T=Ten
; J=Jack
; Q=Queen
; K=King

; So no double use of same capital

field	equ	#4000

dfileok	equ	field-fieldc+dfilec-basic-3
frameok	equ	field-fieldc+framesc-frames
lastkok	equ	field-fieldc+lastkc-lastk

up	equ 	field-fieldc+upc
down	equ 	field-fieldc+downc
left	equ     field-fieldc+leftc
right	equ	field-fieldc+rightc
rnd	equ	field-fieldc+rndc

	org 	#4009

len	equ	21

eoscr	equ	#43ff

; in LOWRES more sysvar are used, but in this way the shortest code
; over sysvar to start machinecode. This saves 11 bytes of BASIC

; DO NOT CHANGE AFTER BASIC+3 (=DFILE)
basic   ld 	h,dfile/256			; highbyte of dfile
        jr 	init1

	db 	236				; BASIC over DFILE data
	db 	212,28				; GOTO USR 0
	db 	126,143,0,18			; short FP notation for #8009, no NEWLINE for BASIC

eline	dw 	last
chadd	dw 	last-1
        db 	0,0,0,0,0,0			; x
berg	db 	0				; x

mem	db 	0,0				; x OVERWRITTEN ON LOAD

init1	ld 	l, dfile mod 256		; low byte of dfile
	jr 	init2			
		
lastk	db 	255,255,255
margin  db 	55

nxtlin  dw 	basic				; BASIC-line over sysvar	

flagx   equ 	init2+2
init2	ld 	(basic+3),hl			; repair correct DFILE flagx will be set with 64, the correct value

	ld	h,eoscr/256
	db 	0,0				; x used by ZX81, not effective code after loading

	ld	l,eoscr mod 256
frames  db 	#37
	db	#e9				; set a end of screen marker

	xor	a
	ex	af,af'
	jp 	init				; initroutine in screenmemory

cdflag  db 	64
; DO NOT CHANGE SYSVAR ABOVE UNLESS YOU KNOW WHAT YOU DO

; deck of cards memory
; d  = deckcard			1=deck 		0=on table
; 
; v  = value of card
; sc = suit of card
; c  = colour of card

;                d vvvvsc
cards	db	%00000100	; ace of clubs
	db	%00001000	; 2 of clubs
	db	%00001100	; 3
	db	%00010000	; 4
	db	%00010100	; 5
	db	%00011000	; 6
	db	%00011100	; 7
	db	%00100000	; 8
	db	%00100100	; 9
	db	%00101000	; 10
	db	%00101100	; jack
	db	%00110000	; queen
	db	%00110100	; king

	db	%00000101	; ace of hearts
	db	%00001001	; 2 of hearts
	db	%00001101	; 3
	db	%00010001	; 4
	db	%00010101	; 5
	db	%00011001	; 6
	db	%00011101	; 7
	db	%00100001	; 8
	db	%00100101	; 9
	db	%00101001	; 10
	db	%00101101	; jack
	db	%00110001	; queen
	db	%00110101	; king

	db	%00000110	; ace of spades
	db	%00001010	; 2 of spades
	db	%00001110	; 3
	db	%00010010	; 4
	db	%00010110	; 5
	db	%00011010	; 6
	db	%00011110	; 7
	db	%00100010	; 8
	db	%00100110	; 9
	db	%00101010	; 10
	db	%00101110	; jack
	db	%00110010	; queen
	db	%00110110	; king

	db	%00000111	; ace of diamonds
	db	%00001011	; 2 of diamonds
	db	%00001111	; 3
	db	%00010011	; 4
	db	%00010111	; 5
	db	%00011011	; 6
	db	%00011111	; 7
	db	%00100011	; 8
	db	%00100111	; 9
	db	%00101011	; 10
	db	%00101111	; jack
	db	%00110011	; queen
	db	%00110111	; king
	db	255

; max 12 cards to move, King to 2, Ace can always go to endpos
	block	12,0		; moving card stack
movest	equ	$-1

; can be redefined at start of the game
keytab	db	%11111011			; Q
	db	1
	db	up mod 256
	db	%11111101			; A
	db	1
	db	down mod 256
	db	%11011111			; O
	db	2
	db	left mod 256
	db	%11011111			; P
	db	1
	db	right mod 256
	db	%11111110			; Z
	db	2	
	db	fire mod 256
	db	%11111011			; R
	db	8
endkey	db	restrt mod 256

; backside, empty field, empty ace and empty top
spcsuit	db	8,14,20,21

; clubs, hearts, spades, diamonds
suittab	db	"C"+101,"H"-27,"S"+101,"D"-27

; Ace to King
valtab	db	"A"-27,30,31,32,33,34,35,36,37
	db	"T"-27,"J"-27,"Q"-27,"K"-27
valpnt	


fire	jp	select		; handle select elsehere

restrt	pop	hl		; drop	return

start	ld	hl,line1
cls	ld	(hl),#88	; set blockfield
sklf	inc	hl
	ld	a,(hl)		; get next field
	cp	#76		; test newline
	jr	z,sklf		; skip new line
	cp	#e9		; test end
	jr	nz,cls		; continue until at end

	pop	hl		; drop cardpointer

	ld	b,52		; 52 cards in the game
	ld	l,cards	mod 256	; HL to start of cards
	ld	(sttop+1),hl	; start of cards on top of screen
	
; shuffle the deck
shuffle	ld	a,(hl)		; current card
	or	h		; make all cards deckcard 
	ld	c,a		; save card in C
	push	hl
	call	rnd		; first cardposition
	add	a,cards mod 256
	ld	l,a
	ld	a,(hl)		; get second card		
	ld	(hl),c		; swap with first card
	pop	hl
	or	h		; make all cards deckcard 
	ld	(hl),a		; swap second card
	inc	hl	
	djnz	shuffle		; repeat for all cards
; now all cards are part of the deck and shuffled

; we now show 7 cardstacks on top
	ld	c,#12		; top line of cards
	ld	l,cards mod 256+21 ; first 21 are cards under shown cards
	ld	e,7		; 7 cards on top
showtop	push	hl		; save pointer
	push	de		; save counter
	call	display-1	; show a card on top
	pop	de		; get counter
	ld	hl,stacks	; point to number above top
	dec	e
	ld	a,l
	add	a,e
	ld	l,a

	ld	a,e
	add	a,28
	ld	(hl),a		; set nr of cards under current card

	pop	hl		; get cardpointer
	inc	hl
	dec	c		
	dec	c		
	dec	c
	jp	p,showtop	; final position is 0, so below 0 is end

; display destinations of suitcards
	ld	bc,#c12		; AT position
shace	ld	a,%00000010	; special ACE pos card
	call	display		; display ace position
	dec	c		; next coordinate
	dec	c
	dec	c
	ld	a,7
	cp	c		; test 4 displays done
	jr	c,shace

	ld	hl,cards+27	; fixed start of deck -1
deckshw	push	hl		; stackposition
	push	bc		; cursorposition

	ld	bc,#c00
	inc	hl
	ld	a,(hl)
	inc	a		; card is backside
	jr	z,deckok	; test end of deck reached
	ld	a,1		; card is empty stack
deckok	call	display		; show card

	pop	bc		; get cursor

curs	xor	a
	in	a,(254)
	and	2		; wait for release select key
	jr	z,curs

cur2	call	field+1
	push	af		; save background
curdisp	ld	(hl),23		; show cursor or selected card
	push	hl

	ld	a,256-5		; delay or game is too fast
	call	delay

	pop	de	
	pop	af
	ld	(de),a		; repair background

	ld	hl,keytab -1	; now handle keyinput
nkey	inc	hl
	ld	a,(hl)		; get inport to read
	inc	hl
	in	a,(254)
	and	(hl)		; test key pressed
	inc	hl
	ld	e,l		; save pointer
	ld	l,(hl)		; preset HL to go to
	ld	a,12		; preload for bottom test
	call	z,#7d		; when keypressed do CALL (HL)
	ld	l,e		; get pointer
	ld	a,l
	sub	endkey mod 256	; last key done?
	jr	nz,nkey		; do other keys
	jr	cur2		; go back up

; we replace the selected card(s) on the board
setcard	ld	(hl),a		; undo cursorset
	ld	hl,movest	; the stack where cards are placed
	push	bc		; save cursorposition
prcard	push	hl		; save cardpointer
	call	display-1	; get card and display card 
	inc	b		; 1 line down
	pop	hl		; get pointer
	dec	hl		; go to next pointer
	ld	a,(hl)		; get value
	inc	a		; test end reached
	jr	nz,prcard	; do other cards
	pop	bc		; get cursor
c2	jr	curs		; back in the game

; rotate the deck back to start
rotdeck	pop	hl		; drop current deckpointer
	ld	hl,cards+27	; go to start of deck-1
fstd	inc	hl		; next card in deck
	bit	6,(hl)		; is it still a deck card?
	jr	z,fstd		; find start of deck
	dec	hl		; start now calculated 
	jr	deck2+1		; do display

; did we select the empty stack?
cardtst	cp	8
	jr	z,rotdeck	; rotate the cards in the deck 

; we selected a possible card, now check if so	
; if already selected then display destination
	ld	d,a		; save background for finding card
	ld	hl,curdisp+1
	
	ld	a,23
	cp	(hl)		; test transport of cards
	jr	nz,setcard	; we print cards on new position
	
	push	hl
	push	de
	call	getfrst		; first card is allowed at bottom
	pop	de
	pop	hl
	jp	m,curs		; no card selected

	ld	(hl),d		; make cursor selected card
	
; we now place cards in movestack, erase old cards and display card below

	push	bc		; save original position
	ld	hl,movest
crdmv	ld	(hl),a		; store latest card
	dec	hl	
	ld	(hl),255	; set endmarker
	inc	b
	push	hl
	call	getcrd		; get next card or test end
	pop	hl
	jp	p,crdmv		; next card to store
	pop	bc

; show card below and erase all under
	ld	a,b
	cp	#c		; are we at the bottom
	call	nz,up		; if not, move 1 position up
	cp	b		; test if we moved up
	jr	z,ficard	; same pos, top or deck 
	call	getcrd		; get card that is here
	jr	hlok	

; we need to find the card below the current card
ficard	cp	#c
	jr	nz,topst	; we need to do a topstacksearch

; deck search
	pop	hl		; get current stackpos

fhl	dec	hl
	ld	a,cards mod 256 +27
	cp	l
	jr	z,dckem		; deck empty
	bit	6,(hl)
	jr	z,fhl
; here find first NEXT card on deck
	ld	a,(hl)

dckem	push	hl
	jr	nz,hlok	
; emptycard		
	xor	a
	call	display
	db	62		; skip POP BC to join 2 routines
curret	pop	bc
c1	jr	c2

select	pop	hl		; drop return
	call	field+1		; get screenvalue
	cp	14		; is it backside, so on the stack?
	jr	nz,cardtst	; not 3 card stackmove

; turn 3 cards from the deck
deck2	pop	hl		; get deckpointer
	ld	d,3		; 3 cards to show
fdeck	inc	hl
	bit	6,(hl)
	jr	z,fdeck		; find card
	ld	a,(hl)
	inc	a
	dec	hl
deck1	jp	z,deckshw	; deck is empty
	inc	hl
	push	bc		; save cursor
	push	de
	ld	bc,#c03		; print besides deck
	push	hl
	call	display-1	; show a card

	ld	a,256-30
	call	delay		; some delay to see cards turned

	pop	hl
	pop	de
	pop	bc
	dec	d
	jr	nz,fdeck	; show 3 cards
	jr	deck1


topst	ld	hl,stacks-1	; pointer for remaining cards on top
	ld	a,c
fsta	inc	hl
	sub	3
	jr 	nc,fsta		; calculate right pointer
	ld	a,(hl)		; get value
	cp	28		; test card remaining
	ld	a,3		; preset empty position
	jr	z,hlok		; show emptycard
	dec	(hl)		; take 1 from stack
sttop	ld	hl,0		; current top stack pointer
	ld	a,(hl)		; get card from top stack
	inc	hl	
	ld	(sttop+1),hl	; set new pointer	

hlok	call	display		; show card under taken card
; now erase until at bottom
	push	bc
	inc	b
	inc	b
eral	inc	b
	ld	a,#c-1
	cp	b
	jr	c,curret	
	call	field+1
	ld	(hl),#88
	inc	hl
	ld	(hl),#88
	jr	eral

; display must stop when over bottom cards
	ld	a,(hl)		; get cardvalue
display	ld	l,a		; temp save
	rra
	rra	
	and	15		; test value
	ld	e,a		; value to E
	ld	a,l
	ld	hl,suittab-1	; normal cards
	jr	nz,nrmcrd	
	ld	l,spcsuit mod 256 -1	; special cards
nrmcrd	and	3		; suit only
	ld	d,a
	inc	d
	push	bc
fsuit	dec	d
	inc	hl
	jr	nz,fsuit
	ld	d,(hl)		; suit in D or special card

	ld	l,valtab mod 256 -1	
fval	inc	hl
	dec	e
	jr	nz,fval

	ld	e,(hl)		; get visible text of value

	rl	e		; e*2
	ld	a,d
	add	a,a		; bit 7 in carry	
	rr	e		; bit 7 now in E

eok	cp	44		; test special cards
	jr	nc,eok1
	ld	e,d		; e and d both same
eok1	call	field+1		; get field
	ld	(hl),e		; show top part of card
	inc	hl
	ld	(hl),d
	call	field		; get field 1 row down		
	ld	a,b
	cp	#c
	jr	z,exl2		; test print bottom reached
	ld	a,d		; get pointer in A
	cp	22		; test backside or clear field
	jr	c,aok		; "+"forms a block
	and	128		; midcard is space inverted or normal

aok	ld	(hl),a		; show midcard
	inc	hl
	ld	(hl),a
	call	field		; get field 1 row down
exl2	ld	a,b
	pop	bc
	cp	#c
	ret	z
	ld	(hl),d		; show bottom of card
	inc	hl
	ld	(hl),e	
	ret

getcrd	ld	a,#c-1
	sub	b
	ret	m		; test not take card below range

getfrst	call	field+1
crdtest	inc	hl
	ld	e,(hl)		; get possible suit
	res	7,a		; make all cards same colour
	ld	d,13		; cardvalue
	ld	hl,valpnt
	call	fval1		; test selected field holds a cardvalue
	ret	m		; no valid card

	ld	a,d
	add	a,a
	add	a,a		; value *4 in table

	ld	d,e
	ld	e,a
	ld	a,d

	ld	l,suittab mod 256 + 4
	ld	d,3
	call	fval1		; now test second field holds a suit
	ret	m		; no valid card

	ld	a,e
	add	a,d		; now cardval built

; find card in deck and take it off
	ld	l,cards mod 256 -1
fposcrd	inc	hl
	ld	e,a
	xor	(hl)
	and	255-192		; take decksignal, test card found
	ld	a,e
	jr	nz,fposcrd
	res	6,(hl)		; signal taken card
	ret

delay	ld	hl,frames
	add	a,(hl)
wfr	cp	(hl)
	jr	nz,wfr
;	ret			; exit through fval1

fval1	dec	hl
	cp	(hl)		; find matching carddisplay
	ret	z
	dec	d
	ret	m
	jr	fval1
				
space	equ	#4400-$-15*len-2-8

	block	space,0
		
; the display file, Code the lines needed.
dfile 	db 	118

stacks	db	"Q"-27,"A"-27,"O"-27,"P"-27,"Z"-27,"R"-27,143	
	db	118

; each needed line has text and a Newline
line1	db	#e9

redef	db	"U"+101,"D"+101,"L"+101,"R"+101,"F"+101,"S"+101

	ld	c,8
setcol	ld	(hl),d		; the routine to set chroma colouring
	inc	hl
	dec	bc
	ld	a,b
	or	c
	jr	nz,setcol
	ret

	block	3,0
	db	118

copyf
init	ld	a,143
	ld	(stacks+6),a
	call	wupk
	ld	a,(hl)
	sub	"Y"-27
	jr	z,stgame
; redefine keys
	xor	a
	ld	(stacks+6),a		; clear "?"

	ld	bc,redef
	ld	hl,stacks
	ld	de,keytab

rloop	ld	a,(bc)
	ld	(hl),a

	exx
	call	wupk
	ld	a,(hl)
	exx
	ld	(hl),a
	ld	a,(lastk)		; get inport	
	ld	(de),a
	inc	de
	in	a,(254)
	cpl
	and	31
	ld	(de),a
	inc	de
	inc	de
	inc	bc
	inc	hl
	ld	a,(hl)
	or	a
	jr	nz,rloop
	jr	init

stgame	
chtest	ld	h,4			; RND highbyte seed
	ld	b,52			; 52 cards in B' for RND
	exx

	ld	bc,#7fef
	in	a,(c)
	inc	a
	jr	z,skipchr		; skip if no chroma


	ld	hl,#c3ff
black	ld	(hl),%00000111		; first we set all characters as black
	dec	hl
	ld	a,h
	cp	#bf
	jr	nz,black

	inc	hl
	ld	d,%01110010		; now we set the needed red cards
	ld	bc,62*8
	call	setcol			

	ld	hl,#c040
	ld	d,%00000100		; now we set the green characters
	ld	c,14*8
	call	setcol

	ld	hl,#c240
	ld	d,%01000100		; and finally the screen character
	call	setcol-2

	
	ld	bc,#7fef
	ld	a,%00100100		; activate chroma colouring
	out	(c),a


skipchr	ld	(line1),a		; erase forced end of screen
	ex	af,af'			; delay intrupts

	ld	hl,(frames)		; part of randomize seed
	ld	(framesc),hl		; keep current values

	ld	sp,dfile+1024
	ld	hl,start
	push	hl			; set H ok for start
	push	hl			; stack startaddress
	ld	hl,fieldc		; now copy over sysvar
	ld	de,field
	ld	bc,vars-fieldc
	ldir

	ld	hl,line1
	ld	de,copyf
	ld	bc,14*len		; built screen
	jp	#a6e			; and start the game

wupk	inc	a
wdwnk	ld	bc,(lastk)
	ld	a,c
	jr	nz,wupk
	inc	a
	jr	z,wdwnk
	jp	#7bd
		
; field is copied over sysvar
fieldc	inc	b
	push	de
	xor	a
	ld	h,a
	ld	d,a
	ld	l,b

	ld	e,b
	add	hl,hl				; hl= 2B
	add	hl,hl				; hl= 4B
	add	hl,de				; hl= 5B
	add	hl,hl				; hl=10B
	
	db	#da				; JP C,NN
dfilec	dw 	dfile				; DFILE, not codeable in lowres
	add	hl,hl				; hl=20B
	add	hl,de				; hl=21B
	ld	e,c
	add	hl,de				; hl=21b+c
	ld	de,line1			
	add	hl,de				; now screen added
	ld	a,(hl)
	pop	de
	ret

upc	dec	b				; move up
	ret	p
downc	cp	b
	ret	z
	inc	b

ltest	ret	p
rightc	ld	a,len-3
	cp	c
	ret	z
	inc	c
	ret
	
lastkc	db	255,255,255		; LASTK
	db	55			; MARGIN

leftc	dec	c
	jr	ltest

rnd1	add	a,e			; add lowbyte framecounter
rnd2	sub	b			; now make value in range needed
	jr	nc,rnd2
	add	a,b			; A = 0 .. 51	
rndc	exx				; entry to shadow, exit to mainreg
	ret	c			;
	db	17
framesc dw	65535			; in effect LD DE,(frames)
	inc	l			; next RND-pointer
	ld	a,(hl)			; get RND-value from ROM
	jr	rnd1	
		
; vars becomes EOSCR with JP (HL)
vars    db	128
last	equ $   	
end           